from setuptools import setup

setup(name='instapy_testers2',
      version='1.0',
      description='Instapy corrected module for the best usefull',
      packages=['instapy_testers2'],
      author_email='lokasan@inbox.ru',
      zip_safe=False)